@@ -22,17 +22,25 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
 
 #include <X11/Xlibint.h>
 #include <X11/extensions/record.h>
+#include <X11/extensions/XShm.h>
+#if HAVE_X11_EXTENSIONS_SHMPROTO_H
 #include <X11/extensions/shmproto.h>
+#elif HAVE_X11_EXTENSIONS_SHMSTR_H
+#include <X11/extensions/shmstr.h>
+#endif
 #include <X11/extensions/Xdamage.h>
 #include <X11/extensions/Xinerama.h>
 #include <X11/extensions/Xrandr.h>
 #include <X11/extensions/Xrender.h>
-#include <X11/extensions/XShm.h>
 #include <X11/Xcursor/Xcursor.h>
 #include <pixman.h>
 
@@ -62,6 +70,8 @@
 #define DBG(x)
 #endif
 
+#define FORCE_FULL_REDRAW 0
+
 struct display {
 	Display *dpy;
 	struct clone *clone;
@@ -127,6 +137,7 @@ struct output {
 
 struct clone {
 	struct clone *next;
+	struct clone *active;
 
 	struct output src, dst;
 
@@ -141,6 +152,7 @@ struct clone {
 struct context {
 	struct display *display;
 	struct clone *clones;
+	struct clone *active;
 	struct pollfd *pfd;
 #define timer pfd[0].fd
 	Display *record;
@@ -278,7 +290,10 @@ static int timerfd(int hz)
 	struct itimerspec it;
 	int fd;
 
+	fd = -1;
+#ifdef CLOCK_MONOTONIC_COARSE
 	fd = timerfd_create(CLOCK_MONOTONIC_COARSE, TFD_NONBLOCK);
+#endif
 	if (fd < 0)
 		fd = timerfd_create(CLOCK_MONOTONIC, TFD_NONBLOCK);
 	if (fd < 0)
@@ -354,7 +369,6 @@ static int add_fd(struct context *ctx, int fd)
 	return 0;
 }
 
-
 static void display_mark_flush(struct display *display)
 {
 	DBG(("%s mark flush (flush=%d)\n",
@@ -382,7 +396,6 @@ static int mode_equal(const XRRModeInfo *a, const XRRModeInfo *b)
 		a->modeFlags == b->modeFlags);
 }
 
-
 static XRRModeInfo *lookup_mode(XRRScreenResources *res, int id)
 {
 	int i;
@@ -397,8 +410,8 @@ static XRRModeInfo *lookup_mode(XRRScreenResources *res, int id)
 
 static int clone_update_modes__randr(struct clone *clone)
 {
-	XRRScreenResources *from_res, *to_res;
-	XRROutputInfo *from_info, *to_info;
+	XRRScreenResources *from_res = NULL, *to_res = NULL;
+	XRROutputInfo *from_info = NULL, *to_info = NULL;
 	int i, j, ret = ENOENT;
 
 	assert(clone->src.rr_output);
@@ -439,6 +452,7 @@ static int clone_update_modes__randr(struct clone *clone)
 	clone->src.rr_crtc = 0;
 
 	/* Create matching modes for the real output on the virtual */
+	XGrabServer(clone->src.dpy);
 	for (i = 0; i < from_info->nmode; i++) {
 		XRRModeInfo *mode, *old;
 		RRMode id;
@@ -479,6 +493,7 @@ static int clone_update_modes__randr(struct clone *clone)
 
 		XRRAddOutputMode(clone->src.dpy, clone->src.rr_output, id);
 	}
+	XUngrabServer(clone->src.dpy);
 	ret = 0;
 
 err:
@@ -497,8 +512,8 @@ err:
 static int clone_update_modes__fixed(struct clone *clone)
 {
 	char mode_name[80];
-	XRRScreenResources *res;
-	XRROutputInfo *info;
+	XRRScreenResources *res = NULL;
+	XRROutputInfo *info = NULL;
 	XRRModeInfo mode;
 	RRMode id;
 	int i, j, ret = ENOENT;
@@ -513,6 +528,8 @@ static int clone_update_modes__fixed(struct clone *clone)
 	if (info == NULL)
 		goto err;
 
+	XGrabServer(clone->src.dpy);
+
 	/* Clear all current UserModes on the output, including any active ones */
 	if (info->crtc) {
 		DBG(("%s(%s-%s): disabling active CRTC\n", __func__,
@@ -547,12 +564,14 @@ static int clone_update_modes__fixed(struct clone *clone)
 
 	XRRAddOutputMode(clone->src.dpy, clone->src.rr_output, id);
 
+	XUngrabServer(clone->src.dpy);
 	ret = 0;
 err:
 	if (info)
 		XRRFreeOutputInfo(info);
 	if (res)
 		XRRFreeScreenResources(res);
+
 	return ret;
 }
 
@@ -564,14 +583,15 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 	XRROutputInfo *output;
 	XRRModeInfo mode;
 	RRMode id;
-	RROutput rr_output;
+	RROutput rr_output = 0;
 	int i;
 
 	DBG(("%s(%d)\n", __func__, nclone));
+	XGrabServer(dpy);
 
 	res = _XRRGetScreenResourcesCurrent(dpy, display->root);
 	if (res == NULL)
-		return 0;
+		goto out;
 
 	sprintf(output_name, "VIRTUAL%d", nclone);
 
@@ -593,7 +613,7 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 
 	DBG(("%s(%s): rr_output=%ld\n", __func__, output_name, (long)rr_output));
 	if (rr_output == 0)
-		return 0;
+		goto out;
 
 	/* Set any mode on the VirtualHead to make the Xserver allocate another */
 	memset(&mode, 0, sizeof(mode));
@@ -609,7 +629,7 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 	/* Force a redetection for the ddx to spot the new outputs */
 	res = XRRGetScreenResources(dpy, display->root);
 	if (res == NULL)
-		return 0;
+		goto out;
 
 	/* Some else may have interrupted us and installed that new mode! */
 	output = XRRGetOutputInfo(dpy, res, rr_output);
@@ -624,6 +644,9 @@ static RROutput claim_virtual(struct display *display, char *output_name, int nc
 	XRRDeleteOutputMode(dpy, rr_output, id);
 	XRRDestroyMode(dpy, id);
 
+out:
+	XUngrabServer(dpy);
+
 	return rr_output;
 }
 
@@ -824,7 +847,7 @@ static void context_update(struct context *ctx)
 			c = XRRGetCrtcInfo(dpy, res, o->crtc);
 		if (c) {
 			DBG(("%s-%s: (x=%d, y=%d, rotation=%d, mode=%ld) -> (x=%d, y=%d, rotation=%d, mode=%ld)\n",
-			     DisplayString(ctx->display->dpy), output->name,
+			     DisplayString(dpy), output->name,
 			     output->x, output->y, output->rotation, output->mode.id,
 			     c->x, c->y, output->rotation, c->mode));
 
@@ -842,7 +865,7 @@ static void context_update(struct context *ctx)
 			XRRFreeCrtcInfo(c);
 		} else {
 			DBG(("%s-%s: (x=%d, y=%d, rotation=%d, mode=%ld) -> off\n",
-			     DisplayString(ctx->display->dpy), output->name,
+			     DisplayString(dpy), output->name,
 			     output->x, output->y, output->rotation, output->mode.id));
 		}
 		output->rr_crtc = o->crtc;
@@ -871,7 +894,7 @@ static void context_update(struct context *ctx)
 	}
 	XRRFreeScreenResources(res);
 
-	DBG(("%s changed? %d\n", DisplayString(ctx->display->dpy), context_changed));
+	DBG(("%s changed? %d\n", DisplayString(dpy), context_changed));
 	if (!context_changed)
 		return;
 
@@ -920,6 +943,7 @@ static void context_update(struct context *ctx)
 		if (res == NULL)
 			continue;
 
+		XGrabServer(display->dpy);
 		for (clone = display->clone; clone; clone = clone->next) {
 			struct output *src = &clone->src;
 			struct output *dst = &clone->dst;
@@ -1005,9 +1029,21 @@ err:
 					 &dst->rr_output, 1);
 			dst->rr_crtc = rr_crtc;
 		}
+		XUngrabServer(display->dpy);
 
 		XRRFreeScreenResources(res);
 	}
+
+	ctx->active = NULL;
+	for (n = 0; n < ctx->nclone; n++) {
+		struct clone *clone = &ctx->clones[n];
+
+		if (clone->dst.rr_crtc == 0)
+			continue;
+
+		clone->active = ctx->active;
+		ctx->active = clone;
+	}
 }
 
 static Cursor display_load_invisible_cursor(struct display *display)
@@ -1060,9 +1096,10 @@ static void display_load_visible_cursor(struct display *display, XFixesCursorIma
 
 	DBG(("%s marking cursor changed\n", DisplayString(display->dpy)));
 	display->cursor_moved++;
-	display->cursor_visible += display->cursor != display->invisible_cursor;
-
-	context_enable_timer(display->ctx);
+	if (display->cursor != display->invisible_cursor) {
+		display->cursor_visible++;
+		context_enable_timer(display->ctx);
+	}
 }
 
 static void display_cursor_move(struct display *display, int x, int y, int visible)
@@ -1096,7 +1133,6 @@ static void display_flush_cursor(struct display *display)
 	}
 
 	XWarpPointer(display->dpy, None, display->root, 0, 0, 0, 0, x, y);
-	display_mark_flush(display);
 
 	cursor = None;
 	if (display->cursor_visible)
@@ -1108,6 +1144,8 @@ static void display_flush_cursor(struct display *display)
 		display->cursor = cursor;
 	}
 
+	display_mark_flush(display);
+
 	display->cursor_moved = 0;
 	display->cursor_visible = 0;
 }
@@ -1316,6 +1354,13 @@ static int clone_paint(struct clone *c)
 	if (c->dst.serial > LastKnownRequestProcessed(c->dst.dpy))
 		return EAGAIN;
 
+	if (FORCE_FULL_REDRAW) {
+		c->damaged.x1 = c->src.x;
+		c->damaged.y1 = c->src.y;
+		c->damaged.x2 = c->src.x + c->width;
+		c->damaged.y2 = c->src.y + c->height;
+	}
+
 	clip.x = c->damaged.x1;
 	clip.y = c->damaged.y1;
 	clip.width  = c->damaged.x2 - c->damaged.x1;
@@ -1352,7 +1397,6 @@ static void usage(const char *arg0)
 static void record_callback(XPointer closure, XRecordInterceptData *data)
 {
 	struct context *ctx = (struct context *)closure;
-	int n;
 
 	DBG(("%s\n", __func__));
 
@@ -1365,8 +1409,10 @@ static void record_callback(XPointer closure, XRecordInterceptData *data)
 
 		if (e->u.u.type == MotionNotify &&
 		    e->u.keyButtonPointer.root == ctx->display->root) {
-			for (n = 0; n < ctx->nclone; n++)
-				clone_move_cursor(&ctx->clones[n],
+			struct clone *clone;
+
+			for (clone = ctx->active; clone; clone = clone->active)
+				clone_move_cursor(clone,
 						  e->u.keyButtonPointer.rootX,
 						  e->u.keyButtonPointer.rootY);
 		}
@@ -1581,6 +1627,7 @@ static int clone_init_depth(struct clone *clone)
 	DBG(("%s-%s wants depth %d\n",
 	     DisplayString(clone->dst.dpy), clone->dst.name, clone->depth));
 
+	ret = -1;
 	for (depth = clone->depth; depth <= 24; depth += 8) {
 		ret = display_init_render(clone->src.display, depth, &clone->src.use_render);
 		if (ret)
@@ -1649,6 +1696,8 @@ static int add_display(struct context *ctx, Display *dpy)
 	display->invisible_cursor = display_load_invisible_cursor(display);
 	display->cursor = None;
 
+	display_cursor_move(display, 0, 0, 0);
+
 	return ConnectionNumber(dpy);
 }
 
@@ -1744,13 +1793,26 @@ static int display_init_damage(struct display *display)
 		return EINVAL;
 	}
 
-	display->damage = XDamageCreate(display->dpy, display->root, XDamageReportRawRectangles);
+	display->damage = XDamageCreate(display->dpy, display->root, XDamageReportBoundingBox);
 	if (display->damage == 0)
 		return EACCES;
 
 	return 0;
 }
 
+static void display_reset_damage(struct display *display)
+{
+	Damage damage;
+
+	damage = XDamageCreate(display->dpy, display->root, XDamageReportBoundingBox);
+	if (damage) {
+		XDamageDestroy(display->dpy, display->damage);
+		display->damage = damage;
+		XFlush(display->dpy);
+		display->flush = 0;
+	}
+}
+
 static void display_init_randr_hpd(struct display *display)
 {
 	int major, minor;
@@ -1889,6 +1951,9 @@ static int last_display_add_clones__xinerama(struct context *ctx)
 		struct clone *clone = add_clone(ctx);
 		RROutput id;
 
+		if (xi[n].width == 0 || xi[n].height == 0)
+			continue;
+
 		clone->depth = 24;
 		clone->next = display->clone;
 		display->clone = clone;
@@ -1925,12 +1990,16 @@ static int last_display_add_clones__xinerama(struct context *ctx)
 		clone->height = xi[n].height;
 		clone->dst.x = xi[n].x_org;
 		clone->dst.y = xi[n].y_org;
+		clone->dst.rr_crtc = -1;
 		ret = clone_update_modes__fixed(clone);
 		if (ret) {
 			fprintf(stderr, "Failed to clone Xinerama screen %d from display \"%s\"\n",
 				n, DisplayString(display->dpy));
 			return ret;
 		}
+
+		clone->active = ctx->active;
+		ctx->active = clone;
 	}
 	XFree(xi);
 	return 0;
@@ -1943,7 +2012,7 @@ static int last_display_add_clones__display(struct context *ctx)
 	Display *dpy = display->dpy;
 	Screen *scr;
 	char buf[80];
-	int n, ret;
+	int ret;
 	RROutput id;
 
 	DBG(("%s(%s)\n", __func__, DisplayString(display->dpy)));
@@ -1964,7 +2033,7 @@ static int last_display_add_clones__display(struct context *ctx)
 		return ret;
 	}
 
-	sprintf(buf, "WHOLE%d", n);
+	sprintf(buf, "WHOLE");
 	ret = clone_output_init(clone, &clone->dst, display, buf, 0);
 	if (ret) {
 		fprintf(stderr, "Failed to add display \"%s\"\n",
@@ -1985,6 +2054,7 @@ static int last_display_add_clones__display(struct context *ctx)
 	clone->height = scr->height;
 	clone->dst.x = 0;
 	clone->dst.y = 0;
+	clone->dst.rr_crtc = -1;
 	ret = clone_update_modes__fixed(clone);
 	if (ret) {
 		fprintf(stderr, "Failed to clone display \"%s\"\n",
@@ -1992,6 +2062,9 @@ static int last_display_add_clones__display(struct context *ctx)
 		return ret;
 	}
 
+	clone->active = ctx->active;
+	ctx->active = clone;
+
 	return 0;
 }
 
@@ -2041,7 +2114,7 @@ static int first_display_wait_for_ack(struct context *ctx, int timeout, int id)
 {
 	struct display *display = ctx->display;
 	struct pollfd pfd;
-	char expect[5];
+	char expect[6]; /* "1234R\0" */
 
 	sprintf(expect, "%04xR", id);
 	DBG(("%s: wait for act '%c%c%c%c%c'\n",
@@ -2314,6 +2387,8 @@ static void context_cleanup(struct context *ctx)
 	if (res == NULL)
 		return;
 
+	XGrabServer(dpy);
+
 	for (i = 0; i < ctx->nclone; i++) {
 		struct clone *clone = &ctx->clones[i];
 		XRROutputInfo *output;
@@ -2346,8 +2421,11 @@ static void context_cleanup(struct context *ctx)
 		}
 	}
 
+	XUngrabServer(dpy);
 	XRRFreeScreenResources(res);
 
+	if (ctx->singleton)
+		XDeleteProperty(dpy, ctx->display->root, ctx->singleton);
 	XCloseDisplay(dpy);
 }
 
@@ -2510,6 +2588,7 @@ int main(int argc, char **argv)
 	while (!done) {
 		XEvent e;
 		int reconfigure = 0;
+		int rr_update = 0;
 
 		DBG(("polling - enable timer? %d, nfd=%d\n", ctx.timer_active, ctx.nfd));
 		ret = poll(ctx.pfd + !ctx.timer_active, ctx.nfd - !ctx.timer_active, -1);
@@ -2520,21 +2599,23 @@ int main(int argc, char **argv)
 
 		DBG(("poll reports %d fd awake\n", ret));
 		if (ctx.pfd[1].revents) {
-			int damaged = 0;
-
 			DBG(("%s woken up\n", DisplayString(ctx.display[0].dpy)));
 			do {
 				XNextEvent(ctx.display->dpy, &e);
 
 				if (e.type == ctx.display->damage_event + XDamageNotify ) {
 					const XDamageNotifyEvent *de = (const XDamageNotifyEvent *)&e;
+					struct clone *clone;
+
 					DBG(("%s damaged: (%d, %d)x(%d, %d)\n",
 					     DisplayString(ctx.display->dpy),
 					     de->area.x, de->area.y, de->area.width, de->area.height));
-					for (i = 0; i < ctx.nclone; i++)
-						clone_damage(&ctx.clones[i], &de->area);
-					context_enable_timer(&ctx);
-					damaged++;
+
+					for (clone = ctx.active; clone; clone = clone->next)
+						clone_damage(clone, &de->area);
+
+					if (ctx.active)
+						context_enable_timer(&ctx);
 				} else if (e.type == ctx.display->xfixes_event + XFixesCursorNotify) {
 					XFixesCursorImage *cur;
 
@@ -2553,7 +2634,6 @@ int main(int argc, char **argv)
 					DBG(("%s screen changed (reconfigure pending? %d)\n",
 					     DisplayString(ctx.display->dpy), reconfigure));
 					reconfigure = 1;
-					context_enable_timer(&ctx);
 				} else if (e.type == PropertyNotify) {
 					XPropertyEvent *pe = (XPropertyEvent *)&e;
 					if (pe->atom == ctx.singleton) {
@@ -2578,8 +2658,6 @@ int main(int argc, char **argv)
 				}
 			} while (XPending(ctx.display->dpy) || poll(&ctx.pfd[1], 1, 0) > 0);
 
-			if (damaged)
-				XDamageSubtract(ctx.display->dpy, ctx.display->damage, None, None);
 			ret--;
 		}
 
@@ -2596,12 +2674,11 @@ int main(int argc, char **argv)
 					XRRNotifyEvent *re = (XRRNotifyEvent *)&e;
 					if (re->subtype == RRNotify_OutputChange) {
 						XRROutputPropertyNotifyEvent *ro = (XRROutputPropertyNotifyEvent *)re;
-						int j;
+						struct clone *clone;
 
-						for (j = 0; j < ctx.nclone; j++) {
-							if (ctx.clones[j].dst.display == &ctx.display[i] &&
-							    ctx.clones[j].dst.rr_output == ro->output)
-								ctx.clones[j].rr_update = 1;
+						for (clone = ctx.display[i].clone; clone; clone = clone->next) {
+							if (clone->dst.rr_output == ro->output)
+								rr_update = clone->rr_update = 1;
 						}
 					}
 				}
@@ -2610,24 +2687,37 @@ int main(int argc, char **argv)
 			ret--;
 		}
 
-		if (reconfigure)
+		if (reconfigure) {
 			context_update(&ctx);
+			display_reset_damage(ctx.display);
+		}
 
-		for (i = 0; i < ctx.nclone; i++)
-			clone_update(&ctx.clones[i]);
+		if (rr_update) {
+			for (i = 0; i < ctx.nclone; i++)
+				clone_update(&ctx.clones[i]);
+		}
 
 		XPending(ctx.record);
 
 		if (ctx.timer_active && read(ctx.timer, &count, sizeof(count)) > 0) {
+			struct clone *clone;
+
 			DBG(("%s timer expired (count=%ld)\n", DisplayString(ctx.display->dpy), (long)count));
 			ret = 0;
 
-			for (i = 0; i < ctx.nclone; i++)
-				ret |= clone_paint(&ctx.clones[i]);
+			if (ctx.active) {
+				DBG(("%s clearing damage\n", DisplayString(ctx.display->dpy)));
+				XDamageSubtract(ctx.display->dpy, ctx.display->damage, None, None);
+				ctx.display->flush = 1;
+			}
+
+			for (clone = ctx.active; clone; clone = clone->active)
+				ret |= clone_paint(clone);
 
 			for (i = 0; i < ctx.ndisplay; i++)
 				display_flush(&ctx.display[i]);
 
+			DBG(("%s timer still active? %d\n", DisplayString(ctx.display->dpy), ret != 0));
 			ctx.timer_active = ret != 0;
 		}
 	}
